@@ -331,8 +331,8 @@ static URI_INLINE UriBool URI_FUNC(FixPercentEncodingMalloc)(const URI_CHAR ** f
 	URI_CHAR * buffer;
 
 	/* Death checks */
-	if ((first == NULL) || (afterLast == NULL) || (*first == NULL)
-		|| (*afterLast == NULL)) {
+	if ((first == NULL) || (afterLast == NULL)
+			|| (*first == NULL) || (*afterLast == NULL)) {
 		return URI_FALSE;
 	}
 
@@ -370,7 +370,7 @@ static URI_INLINE UriBool URI_FUNC(MakeRangeOwner)(unsigned int * doneMask,
 		if (dup == NULL) {
 			return URI_FALSE; /* Raises malloc error */
 		}
-		memcpy(dup, range->afterLast, lenInBytes);
+		memcpy(dup, range->first, lenInBytes);
 		range->first = dup;
 		range->afterLast = dup + lenInChars;
 		*doneMask |= maskTest;
@@ -450,9 +450,9 @@ static URI_INLINE UriBool URI_FUNC(MakeOwner)(URI_TYPE(Uri) * uri,
 	}
 
 	/* Port text, must come last so we don't have to undo that one if it fails. *
-	 * Otherwise we would need and extra enum flag for it althoug the port      *
+	 * Otherwise we would need and extra enum flag for it although the port      *
 	 * cannot go unnormalized...                                                */
-	if (!URI_FUNC(MakeRangeOwner)(doneMask, 0, &(uri->userInfo))) {
+	if (!URI_FUNC(MakeRangeOwner)(doneMask, 0, &(uri->portText))) {
 		return URI_FALSE; /* Raises malloc error */
 	}
 
@@ -463,7 +463,15 @@ static URI_INLINE UriBool URI_FUNC(MakeOwner)(URI_TYPE(Uri) * uri,
 
 unsigned int URI_FUNC(NormalizeSyntaxMaskRequired)(const URI_TYPE(Uri) * uri) {
 	unsigned int res;
+#if defined(__GNUC__) && ((__GNUC__ > 4) \
+        || ((__GNUC__ == 4) && defined(__GNUC_MINOR__) && (__GNUC_MINOR__ >= 2)))
+    /* Slower code that fixes a warning, not sure if this is a smart idea */
+	URI_TYPE(Uri) writeableClone;
+	memcpy(&writeableClone, uri, 1 * sizeof(URI_TYPE(Uri)));
+	URI_FUNC(NormalizeSyntaxEngine)(&writeableClone, 0, &res);
+#else
 	URI_FUNC(NormalizeSyntaxEngine)((URI_TYPE(Uri) *)uri, 0, &res);
+#endif
 	return res;
 }
 
@@ -498,23 +506,26 @@ static URI_INLINE int URI_FUNC(NormalizeSyntaxEngine)(URI_TYPE(Uri) * uri, unsig
 	} else if (inMask == URI_NORMALIZED) {
 		/* Nothing to do */
 		return URI_SUCCESS;
-	} else if ((uri->scheme.first == NULL) && !uri->absolutePath) {
-		/* Normalization only fone for absolute paths */
-		return URI_SUCCESS;
 	}
 
 	/* Scheme, host */
 	if (outMask != NULL) {
 		const UriBool normalizeScheme = URI_FUNC(ContainsUppercaseLetters)(
 				uri->scheme.first, uri->scheme.afterLast);
-		const UriBool normalizeHost = URI_FUNC(ContainsUppercaseLetters)(
+		const UriBool normalizeHostCase = URI_FUNC(ContainsUppercaseLetters)(
 			uri->hostText.first, uri->hostText.afterLast);
 		if (normalizeScheme) {
 			*outMask |= URI_NORMALIZE_SCHEME;
 		}
 
-		if (normalizeHost) {
+		if (normalizeHostCase) {
 			*outMask |= URI_NORMALIZE_HOST;
+		} else {
+			const UriBool normalizeHostPrecent = URI_FUNC(ContainsUglyPercentEncoding)(
+					uri->hostText.first, uri->hostText.afterLast);
+			if (normalizeHostPrecent) {
+				*outMask |= URI_NORMALIZE_HOST;
+			}
 		}
 	} else {
 		/* Scheme */
@@ -552,16 +563,20 @@ static URI_INLINE int URI_FUNC(NormalizeSyntaxEngine)(URI_TYPE(Uri) * uri, unsig
 					&& (uri->hostData.ip6 == NULL)) {
 				/* Regname */
 				if (uri->owner) {
-					URI_FUNC(LowercaseInplace)(uri->hostText.first,
-							uri->hostText.afterLast);
+					URI_FUNC(FixPercentEncodingInplace)(uri->hostText.first,
+							&(uri->hostText.afterLast));
 				} else {
-					if (!URI_FUNC(LowercaseMalloc)(&(uri->hostText.first),
+					if (!URI_FUNC(FixPercentEncodingMalloc)(
+							&(uri->hostText.first),
 							&(uri->hostText.afterLast))) {
 						URI_FUNC(PreventLeakage)(uri, doneMask);
 						return URI_ERROR_MALLOC;
 					}
 					doneMask |= URI_NORMALIZE_HOST;
 				}
+
+				URI_FUNC(LowercaseInplace)(uri->hostText.first,
+						uri->hostText.afterLast);
 			}
 		}
 	}
@@ -612,14 +627,10 @@ static URI_INLINE int URI_FUNC(NormalizeSyntaxEngine)(URI_TYPE(Uri) * uri, unsig
 			}
 			walker = walker->next;
 		}
-	} else {
-		const URI_TYPE(PathSegment) * walker;
-
-		/* 6.2.2.3 Path Segment Normalization */
-		if (!URI_FUNC(RemoveDotSegments)(uri)) {
-			URI_FUNC(PreventLeakage)(uri, doneMask);
-			return URI_ERROR_MALLOC;
-		}
+	} else if (inMask & URI_NORMALIZE_PATH) {
+		URI_TYPE(PathSegment) * walker;
+		const UriBool relative = ((uri->scheme.first == NULL)
+				&& !uri->absolutePath) ? URI_TRUE : URI_FALSE;
 
 		/* Fix percent-encoding for each segment */
 		walker = uri->pathHead;
@@ -639,6 +650,12 @@ static URI_INLINE int URI_FUNC(NormalizeSyntaxEngine)(URI_TYPE(Uri) * uri, unsig
 			}
 			doneMask |= URI_NORMALIZE_PATH;
 		}
+
+		/* 6.2.2.3 Path Segment Normalization */
+		if (!URI_FUNC(RemoveDotSegments)(uri, relative)) {
+			URI_FUNC(PreventLeakage)(uri, doneMask);
+			return URI_ERROR_MALLOC;
+		}
 	}
 
 	/* Query, fragment */
